import itertools as it
import json
import operator as op
import re
import sys
import typing as t
from datetime import datetime, timedelta
from functools import singledispatch

from rich import box
from rich.align import Align
from rich.bar import Bar
from rich.columns import Columns
from rich.console import ConsoleRenderable, Group
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table

from rich_tables.generic import flexitable
from rich_tables.music import albums_table
from rich_tables.utils import (
    FIELDS_MAP,
    border_panel,
    colored_with_bg,
    format_with_color,
    get_val,
    make_console,
    make_difftext,
    md_panel,
    new_table,
    new_tree,
    predictably_random_color,
    progress_bar,
    simple_panel,
    time2human,
    wrap,
)

JSONDict = t.Dict[str, t.Any]


console = make_console()
print = console.print


def pulls_table(data: t.List[JSONDict]) -> t.Iterable[t.Union[str, ConsoleRenderable]]:
    def res_border_style(resolved: bool, outdated: bool) -> str:
        return "green" if resolved else "yellow" if resolved is False else ""

    def comment_panel(comment: JSONDict, **kwargs: t.Any) -> Panel:
        reactions = [
            wrap(f":{r['content'].lower()}:", "bold") + " " + get_val(r, "user")
            for r in comment.get("reactions", [])
        ]
        text = comment["body"]
        split = re.split(r"(?=```.*)```", text)
        rends = []
        for idx, content in enumerate(split):
            if content:
                if idx % 2 == 0:
                    content = md_panel(content)
                else:
                    lang, codeblock = content.split("\n", 1)
                    content = FIELDS_MAP[lang or "python"](text)
                rends.append(content)
        return simple_panel(
            Group(*rends),
            **{
                "title": " ".join(get_val(comment, f) for f in ["author", "createdAt"]),
                "subtitle": "\n".join(reactions) + "\n",
                **kwargs,
            },
        )

    def fmt_add_del(file: JSONDict) -> t.List[str]:
        added, deleted = file["additions"], file["deletions"]
        additions = f"+{added}" if added else ""
        deletions = f"-{deleted}" if deleted else ""
        return [wrap(additions.rjust(5), "b green"), wrap(deletions.rjust(3), "b red")]

    def state_color(state: str) -> str:
        return {
            "True": "green",
            True: "green",
            "APPROVED": "green",
            "RESOLVED": "s green",
            "OPEN": "green",
            "MERGED": "magenta",
            "PENDING": "yellow",
            "OUTDATED": "yellow",
            "COMMENTED": "yellow",
            "CHANGES_REQUESTED": "yellow",
            "REVIEW_REQUIRED": "red",
            "DISMISSED": "gray42",
            "False": "red",
        }.get(state, "default")

    def fmt_state(state: str) -> str:
        return wrap(state, state_color(state))

    exclude = {
        "title",
        "body",
        "reviews",
        "comments",
        "reviewThreads",
        "url",
        "files",
        "commits",
        "isReadByViewer",
        "repository",
        "labels",
        "reviewDecision",
        "state",
        "reviewRequests",
    }
    FIELDS_MAP.update(
        {
            "state": lambda x: wrap(fmt_state(x), "b"),
            "reviewDecision": lambda x: wrap(fmt_state(x), "b"),
            "dates": lambda x: new_table(
                rows=[
                    [wrap(r" ⬤ ", "b green"), time2human(x[0])],
                    [wrap(r" ◯ ", "b yellow"), time2human(x[1])],
                ]
            ),
            "repository": lambda x: x.get("name"),
            "path": lambda x: wrap(x, "b"),
            "message": lambda x: wrap(x, "i"),
            "files": lambda x: border_panel(
                new_table(
                    rows=[
                        [*fmt_add_del(y), get_val(y, "path")]
                        for y in sorted(
                            x,
                            key=lambda x: x["additions"] + x["deletions"],
                            reverse=True,
                        )
                    ]
                ),
                title="files",
                border_style=f"dim {predictably_random_color('files')}",
            ),
            "commits": lambda x: border_panel(
                new_table(
                    rows=[
                        [
                            *fmt_add_del(y),
                            get_val(y, "message"),
                            get_val(y, "committedDate"),
                        ]
                        for y in x
                    ]
                ),
                title="commits",
                border_style=f"dim {predictably_random_color('commits')}",
            ),
            "reviewRequests": lambda x: "  ".join(map(colored_with_bg, x)),
        }
    )

    pr = data[0]
    if pr and "additions" in pr:
        pr["files"].append(
            dict(additions=pr.pop("additions", ""), deletions=pr.pop("deletions", ""))
        )
    pr["dates"] = {"created": pr.pop("createdAt"), "updated": pr.pop("updatedAt")}

    title, name = pr["title"], pr["repository"]["name"]
    repo_color = predictably_random_color(name)
    decision_color = state_color(pr["reviewDecision"])
    keys = sorted(set(pr) - exclude)
    yield border_panel(
        new_table(
            rows=[
                [Align.center(wrap(title, state_color(pr["state"])))],
                [Align.center(get_val(pr, "labels"), vertical="middle")],
                [flexitable([{k: v for k, v in pr.items() if k in keys}])]
                # [
                #     Columns(
                #         map(
                #             lambda x: simple_panel(
                #                 get_val(pr, x),
                #                 title=wrap(x, "b"),
                #                 title_align="center",
                #                 expand=True,
                #                 align="center",
                #             ),
                #             keys,
                #         ),
                #         align="center",
                #         expand=True,
                #         equal=True,
                #     )
                # ],
            ]
        ),
        title=wrap(name, f"b {repo_color}"),
        box=box.DOUBLE_EDGE,
        border_style=decision_color,
        subtitle=f"[b][{decision_color}]{pr['reviewDecision']}[/] [#ffffff]//[/] {fmt_state(pr['state'])}[/]",
        expand=False,
        align="center",
        title_align="center",
        subtitle_align="center",
    )

    yield md_panel(pr["body"])
    yield new_table(rows=[[get_val(pr, "files"), get_val(pr, "commits")]])

    global_comments: t.List[ConsoleRenderable] = []
    raw_global_comments = pr["comments"] + pr["reviews"]
    for comment in sorted(raw_global_comments, key=op.itemgetter("createdAt")):
        state = comment.get("state", "COMMENTED")
        subtitle = "comment"
        if state != "COMMENTED":
            subtitle = "review - " + wrap(state, f"b {state_color(state)}")
        if state != "COMMENTED" or comment["body"]:
            global_comments.append(
                comment_panel(
                    comment,
                    subtitle=subtitle,
                    border_style=predictably_random_color(comment["author"]),
                    box=box.HEAVY,
                )
            )
    if global_comments:
        yield border_panel(
            new_table(rows=[[x] for x in global_comments]), title="Reviews & Comments"
        )

    yield new_table(rows=[[get_val(pr, "reviewRequests")]])
    total_threads = len(pr["reviewThreads"])
    if not total_threads:
        return

    resolved_threads = len(list(filter(lambda x: x["isResolved"], pr["reviewThreads"])))
    table = new_table()
    table.add_row(
        border_panel(
            progress_bar(resolved_threads, total_threads),
            title=f"{resolved_threads} / {total_threads} resolved",
            border_style="dim yellow",
        )
    )

    for thread in pr["reviewThreads"]:
        files: t.List[ConsoleRenderable] = []
        for diff_hunk, comments in it.groupby(
            sorted(
                thread["comments"],
                key=lambda x: (x.get("diffHunk", ""), x.get("createdAt", "")),
            ),
            lambda x: x.get("diffHunk") or "",
        ):
            rows = it.chain.from_iterable(
                [[x], [""]] for x in map(comment_panel, comments)
            )
            comments_col = new_table(rows=rows)
            diff = Syntax(
                diff_hunk,
                "diff",
                theme="paraiso-dark",
                background_color="black",
                # word_wrap=True,
            )
            files.append(
                new_table(rows=[[diff, simple_panel(comments_col)]], highlight=False)
            )

        table.add_row(
            border_panel(
                new_table(rows=it.zip_longest(*(iter(files),) * 2)),
                highlight=False,
                border_style=res_border_style(
                    thread["isResolved"], thread["isOutdated"]
                ),
                title=wrap(thread["path"], "b magenta")
                + " "
                + fmt_state("RESOLVED" if thread["isResolved"] else "PENDING"),
                subtitle=fmt_state("OUTDATED") if thread["isOutdated"] else "",
            )
        )
    yield border_panel(table)


def lights_table(lights: t.List[JSONDict]) -> Table:
    from rgbxy import Converter

    headers = lights[0].keys()
    table = new_table(*headers)
    conv = Converter()
    for light in lights:
        xy = light.get("xy")
        style = ""
        if not light["on"]:
            style = "dim"
            light["xy"] = ""
        elif xy:
            color = conv.xy_to_hex(*xy)
            light["xy"] = wrap("   a", f"#{color} on #{color}")
        table.add_row(*map(str, map(lambda x: light.get(x, ""), headers)), style=style)
    yield table


def calendar_table(events: t.List[JSONDict]) -> t.Iterable[ConsoleRenderable]:
    def get_start_end(start: datetime, end: datetime) -> t.Tuple[int, int]:
        if start.hour == end.hour == 0:
            return 0, 86400
        day_start_ts = start.replace(hour=0).timestamp()
        return int(start.timestamp() - day_start_ts), int(
            end.timestamp() - day_start_ts
        )

    status_map = dict(
        needsAction="[b grey3] ? [/]",
        accepted="[b green] ✔ [/]",
        declined="[b red] ✖ [/]",
        tentative="[b yellow] ? [/]",
    )

    cal_to_color = {e["calendar"]: e["backgroundColor"] for e in events}
    if len(cal_to_color) > 1:
        color_key = "calendar"
        color_id_to_color = cal_to_color
    else:
        color_key = "colorId"
        color_id_to_color = {
            e[color_key]: predictably_random_color(e[color_key]) for e in events
        }
    for e in events:
        e["color"] = color_id_to_color[e[color_key]]
    cal_fmted = [wrap(f" {c} ", f"b black on {clr}") for c, clr in cal_to_color.items()]
    yield Columns(cal_fmted, expand=True, equal=False, align="center")

    new_events: t.List[JSONDict] = []
    for event in events:
        start_iso, end_iso = event["start"], event["end"]
        orig_start = datetime.fromisoformat(
            (start_iso.get("dateTime") or start_iso.get("date")).strip("Z")
        )
        orig_end = datetime.fromisoformat(
            (end_iso.get("dateTime") or end_iso.get("date")).strip("Z")
        )
        h_after_midnight = (
            24 * (orig_end - orig_start).days
            + ((orig_end - orig_start).seconds // 3600)
        ) - (24 - orig_start.hour)

        end = (orig_start + timedelta(days=1)).replace(hour=0, minute=0, second=0)

        def eod(day_offset: int) -> datetime:
            return (orig_start + timedelta(days=day_offset)).replace(
                hour=23, minute=59, second=59
            )

        def midnight(day_offset: int) -> datetime:
            return (orig_start + timedelta(days=day_offset)).replace(
                hour=0, minute=0, second=0
            )

        days_count = h_after_midnight // 24 + 1
        for start, end in zip(
            [orig_start, *map(midnight, range(1, days_count + 1))],
            [*map(eod, range(days_count)), orig_end],
        ):
            if start == end:
                continue

            color = (
                "grey7" if end.replace(tzinfo=None) < datetime.now() else event["color"]
            )
            new_events.append(
                {
                    **event,
                    **dict(
                        color=color,
                        summary=status_map[event["status"]]
                        + wrap(event["summary"] or "busy", f"b {color}"),
                        start=start,
                        start_day=start.strftime("%d %a"),
                        start_time=wrap(start.strftime("%H:%M"), "white"),
                        end_time=wrap(end.strftime("%H:%M"), "white"),
                        desc=border_panel(get_val(event, "desc"))
                        if event["desc"]
                        else "",
                        bar=Bar(86400, *get_start_end(start, end), color=color),
                    ),
                }
            )

    keys = "summary", "start_time", "end_time", "bar"
    month_events: t.Iterable[JSONDict]
    for month_tuple, month_events in it.groupby(
        new_events, lambda x: (x["start"].month, x["start"].strftime("%Y %B"))
    ):
        month_events = sorted(month_events, key=lambda x: x.get("start_day") or "")
        _, year_and_month = month_tuple

        table = new_table(*keys, highlight=False, padding=0, show_header=False)
        for day, day_events in it.groupby(
            month_events, lambda x: x.get("start_day") or ""
        ):
            table.add_row(wrap(day, "b i"))
            for event in day_events:
                if "Week " in event["summary"]:
                    table.add_row("")
                    table.add_dict_item(event, style=event["color"] + " on grey7")
                else:
                    table.add_dict_item(event)
                    if event["desc"]:
                        table.add_row("", "", "", event["desc"])
            table.add_row("")
        yield border_panel(table, title=year_and_month)


def tasktime(datestr: str):
    return time2human(datetime.strptime(datestr, "%Y%m%dT%H%M%SZ").timestamp())


def tasks_table(tasks: t.List[JSONDict]) -> t.Iterator:
    fields_map: JSONDict = dict(
        id=str,
        urgency=lambda x: str(round(x, 1)),
        description=lambda x: x,
        due=tasktime,
        end=tasktime,
        sched=tasktime,
        tags=lambda x: " ".join(map(format_with_color, x or [])),
        project=format_with_color,
        modified=tasktime,
        annotations=lambda l: "\n".join(
            map(
                lambda x: wrap(tasktime(x["entry"]), "b")
                + ": "
                + wrap(x["description"], "i"),
                l,
            )
        ),
    )
    FIELDS_MAP.update(fields_map)
    status_map = {
        "completed": "b s black on green",
        "deleted": "s red",
        "pending": "white",
        "started": "b green",
        "recurring": "i magenta",
    }
    id_to_desc = uuid_to_id = {}
    for task in tasks:
        task["sched"] = task.get("scheduled")
        if not task.get("id"):
            task["id"] = task["uuid"].split("-")[0]
        uuid_to_id[task["uuid"]] = task["id"]

        if task.get("start"):
            task["status"] = "started"

        recur = task.get("recur")
        if recur:
            if task.get("status") == "recurring":
                continue
            else:
                task["status"] = "recurring"
                task["description"] += f" ({recur})"
        desc = wrap(task["description"], status_map[task["status"]])
        if task.get("priority") == "H":
            desc = f"[b]([red]![/])[/] {desc}"
        id_to_desc[task["id"]] = desc

    group_by = tasks[0].get("group_by") or ""
    headers = ["urgency", "id"] + [
        k for k in fields_map.keys() if k not in {group_by, "id", "urgency"}
    ]

    for group, task_group in it.groupby(
        sorted(
            tasks,
            key=lambda x: (x.get(group_by) or "", x.get("urgency") or 0),
            reverse=True,
        ),
        lambda x: x.get(group_by) or f"no [b]{group_by}[/]",
    ):
        project_color = predictably_random_color(str(group))
        table = new_table(padding=(0, 1, 0, 1))
        for task in task_group:
            task_obj = id_to_desc.get(task["id"])
            if not task_obj:
                continue

            task_tree = new_tree(title=task_obj, guide_style=project_color)
            ann = task.pop("annotations", None)
            if ann:
                task_tree.add(FIELDS_MAP["annotations"](ann))
            for kuid in task.get("depends") or []:
                dep = id_to_desc.get(uuid_to_id.get(kuid))
                if dep:
                    task_tree.add(wrap(str(uuid_to_id[kuid]), "b") + " " + dep)
            task["description"] = task_tree
            table.add_row(*map(lambda x: get_val(task, x), headers))

        for col in table.columns.copy():
            try:
                next(filter(op.truth, col.cells))
            except StopIteration:
                table.columns.remove(col)
        panel = simple_panel if group == "started" else border_panel
        yield panel(table, title=wrap(group, "b"), style=project_color)


def load_data() -> t.Any:
    text = re.sub(r"\x00", "", sys.stdin.read())
    try:
        data = json.loads(text)
        assert data and (data.get("values") if "values" in data else True)
    except (json.JSONDecodeError, AssertionError):
        console.log(flexitable(text))
        # console.log(wrap("No data", "b red"), log_locals=True)
        exit(1)
    else:
        return data


@singledispatch
def draw_data(data: t.Union[JSONDict, t.List], title: str = "") -> None:
    pass


@draw_data.register(dict)
def _draw_data_dict(data: JSONDict) -> t.Iterator:
    if "values" in data and "title" in data:
        values, title = data["values"], data["title"]
        calls: t.Dict[str, t.Callable] = {
            "Pull Requests": pulls_table,
            "Hue lights": lights_table,
            "Calendar": calendar_table,
            "Album": albums_table,
            "Tasks": tasks_table,
        }
        if title in calls:
            yield from calls[title](values)
        else:
            yield flexitable(values)
    else:
        yield flexitable(data)


@draw_data.register(list)
def _draw_data_list(data: t.List[JSONDict], title: str = "") -> t.Iterator:
    yield flexitable(data)


def main():
    args = []
    if len(sys.argv) > 1:
        args.extend(sys.argv[1:])

    if args and args[0] == "diff":
        console.print(make_difftext(*args[1:]), highlight=False)
        return

    if "-s" in set(args):
        console.record = True

    data = load_data()
    if "-j" in args:
        console.print_json(data=data)
    else:
        try:
            for ret in draw_data(data):
                console.print(ret)
        except Exception:
            console.print_exception(show_locals=True)

    if "-s" in set(args):
        console.save_html("saved.html")


if __name__ == "__main__":
    main()



import itertools as it
import operator as op
import re
from collections import defaultdict
from functools import lru_cache, partial
from typing import Any, Callable, Dict, Iterable, List, Tuple

from rich import box
from rich.align import Align
from rich.console import ConsoleRenderable, Group
from rich.panel import Panel
from rich.table import Table

from rich_tables.utils import (
    DISPLAY_HEADER,
    FIELDS_MAP,
    border_panel,
    new_table,
    predictably_random_color,
    simple_panel,
    wrap,
)

JSONDict = Dict[str, Any]

TRACK_FIELDS = [
    "track",
    "length",
    "artist",
    "title",
    "bpm",
    "last_played",
    "plays",
    "skips",
    "helicopta",
    "hidden",
]
ALBUM_IGNORE = set(TRACK_FIELDS) | {
    "album_color",
    "albumartist_color",
    "album",
    "album_title",
    "comments",
    "genre",
    "tracktotal",
    "albumartist",
}


new_table = partial(new_table, collapse_padding=True, expand=True, padding=0)


def get_header(key: str) -> str:
    return DISPLAY_HEADER.get(key, key)


def get_def(obj: JSONDict, default: Any = "") -> Callable[[str], Any]:
    def get_value(key: str) -> Any:
        return obj.get(key) or default

    return get_value


@lru_cache(maxsize=128)
def get_val(track: JSONDict, field: str) -> Any:
    trackdict = dict(track)
    return FIELDS_MAP[field](trackdict[field]) if trackdict.get(field) else ""


def get_vals(
    fields: Iterable[str], tracks: Iterable[JSONDict]
) -> Iterable[Iterable[str]]:
    return [[get_val(tuple(t.items()), f) for f in fields] for t in tracks]


def simple_tracks_table(tracks, fields, color, sort):
    # type: (List[JSONDict], Iterable[str], str, bool) -> Table
    return new_table(
        rows=get_vals(
            fields,
            sorted(tracks, key=op.methodcaller("get", "track", "")) if sort else tracks,
        ),
        expand=False,
    )


def _tracks_table(tracks, fields, color, sort):
    # type: (List[JSONDict], Iterable[str], str, bool) -> Table
    return new_table(
        *map(get_header, fields),
        rows=get_vals(
            fields,
            sorted(tracks, key=op.methodcaller("get", "track", "")) if sort else tracks,
        ),
        border_style=color,
        padding=(0, 0, 0, 1),
    )


def album_stats(tracks: List[JSONDict]) -> JSONDict:
    def agg(field: str, default: int = 0) -> Iterable[int]:
        return ((x.get(field) or default) for x in tracks)

    stats: JSONDict = dict(
        bpm=round(sum(agg("bpm")) / len(tracks)),
        rating=round(sum(agg("rating")) / len(tracks), 2),
        plays=sum(agg("plays")),
        skips=sum(agg("skips")),
        mtime=max(agg("mtime")),
        last_played=max(agg("last_played")),
        tracktotal=(str(len(tracks)), str(tracks[0].get("tracktotal")) or str(0)),
        comments="\n---\n---\n".join(set(agg("comments", ""))),
    )
    return stats


def add_colors(album: JSONDict) -> None:
    for field in "album", "albumartist":
        val = (album.get(field) or "").replace("Various Artists", "VA")
        color = predictably_random_color(val)
        album[f"{field}_color"] = color
        val = album.get(field)
        album[field] = wrap(val, f"b i {color}") if val else ""


def format_title(title: str) -> str:
    return wrap(f"  {title}  ", "i white on grey3")


def album_title(album: JSONDict) -> Table:
    name = album["album"]
    artist = album.get("albumartist") or album.get("artist")
    genre = album.get("genre") or ""
    released = album.get("released", "")
    return new_table(
        rows=[[format_title(f"{name} by {artist}"), format_title(released), genre]]
    )


def album_info(tracks: List[JSONDict]) -> JSONDict:
    first = tracks[0]
    fields = sorted([f for f in tracks[0] if f not in TRACK_FIELDS])

    album = defaultdict(str, zip(fields, op.itemgetter(*fields)(first)))
    album.update(**album_stats(tracks))
    add_colors(album)
    for field, val in filter(op.truth, sorted(album.items())):
        album[field] = get_val(tuple(album.items()), field)
    album["album_title"] = album_title(album)
    return album


def album_info_table(album: JSONDict) -> Table:
    def should_display(keyval: Tuple[str, Any]) -> bool:
        return keyval[1] and keyval[0] not in ALBUM_IGNORE

    items = filter(should_display, sorted(album.items()))
    table = new_table(rows=map(lambda x: (get_header(x[0]), x[1]), items))
    table.columns[0].style = "b " + album["album_color"]
    return table


def simple_album_panel(tracks: List[JSONDict]) -> Panel:
    album = album_info(tracks)

    get = album.get

    albumtype = get_val(album, "albumtypes")
    title = ""
    name = get("album")
    if name:
        label = get("label")
        title = format_title(
            (f"{label}: " if label else "")
            + " by ".join(filter(op.truth, [name, get("albumartist") or ""]))
            + f" ({albumtype})"
        )
        fields = TRACK_FIELDS
    else:
        fields = dict.fromkeys(k for k in tracks[0] if k != "albumtypes")

    color = predictably_random_color(str(len(tracks)))
    tracklist = simple_tracks_table(tracks, fields, album["album_color"], sort=False)
    return border_panel(tracklist, title=title, style=color)


def detailed_album_panel(tracks: List[JSONDict]) -> Panel:
    album = album_info(tracks)

    # ignore the artist field if there is only one found
    track_fields = [
        t
        for t in TRACK_FIELDS
        if len(tracks) > 1 or len(set(map(lambda x: x.get("artist"), tracks))) == 1
    ]
    tracklist = _tracks_table(tracks, track_fields, album["album_color"], sort=True)

    _, track = max(map(op.itemgetter("last_played", "track"), tracks))
    if int(re.sub(r"\D", "", str(track).replace("A", "1"))) > 0:
        row_no = tracklist.columns[0]._cells.index(str(track))
        tracklist.rows[row_no].style = "b white on #000000"
        tracklist.add_row(
            *[album.get(k) or "" for k in ["tracktotal", *track_fields[1:]]],
            style="d white on grey11",
        )

    comments = album.get("comments")
    return border_panel(
        Group(
            album["album_title"],
            Align.center(
                simple_panel(comments, style="grey54", expand=True, align="center")
            )
            if comments
            else "",
            new_table(rows=[[album_info_table(album), tracklist]]),
        ),
        box=box.DOUBLE_EDGE,
        style=album["albumartist_color"],
    )


def albums_table(all_tracks: List[JSONDict]) -> Iterable[ConsoleRenderable]:
    def get_album(track: JSONDict) -> str:
        return track.get("album") or ""

    for track in all_tracks:
        if not track["album"] and "single" in track.get("albumtype", ""):
            track["album"] = "singles"
            track["albumartist"] = track["label"]

    for album_name, tracks in it.groupby(sorted(all_tracks, key=get_album), get_album):
        yield detailed_album_panel(list(tracks))



import itertools as it
import json
from functools import partial
from typing import Any, Dict, Iterable, List, Optional, Union

from multimethod import multimethod
from rich import box
from rich.columns import Columns
from rich.console import ConsoleRenderable, RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text

from .utils import (
    DISPLAY_HEADER,
    FIELDS_MAP,
    NewTable,
    border_panel,
    counts_table,
    format_with_color,
    make_console,
    make_difftext,
    new_table,
    predictably_random_color,
    simple_panel,
    wrap,
)

JSONDict = Dict[str, Any]
console = make_console()


def mapping_view_table(**kwargs: Any) -> NewTable:
    """A table with two columns
    * First for bold field names
    * Second one for values
    """
    table = new_table(border_style="misty_rose1", box=box.MINIMAL, expand=False)
    table.add_column(justify="right", style="bold misty_rose1")
    table.add_column()
    return table


def prepare_dict(item: JSONDict) -> JSONDict:
    if "before" in item and "after" in item:
        before, after = item.pop("before"), item.pop("after")
        if not item.get("diff"):
            if isinstance(before, list):
                before, after = "\n".join(before), "\n".join(after)

            if isinstance(before, str):
                item["diff"] = make_difftext(before, after)
            else:
                keys = before.keys()
                item["diff"] = json.dumps(
                    {
                        k: make_difftext(str(before[k] or ""), str(after[k] or ""))
                        for k in keys
                    },
                    indent=2,
                )
    return item


@multimethod
def flexitable(data: None, header: str = "") -> RenderableType:
    return str(data)


@flexitable.register
def _(data: str) -> RenderableType:
    if "[/]" not in data:
        data = data.replace("[", "").replace("]", "")
    return " | ".join(map(format_with_color, data.split(" | ")))


@flexitable.register
def _(data: str, header: str) -> RenderableType:
    if "[/]" not in data:
        data = data.replace("[", "").replace("]", "")
    return FIELDS_MAP[header](data)


@flexitable.register
def _(data: Union[int, float], header: Optional[str] = "") -> RenderableType:
    return flexitable(str(data), header)


@flexitable.register
def _(data: JSONDict, header: Optional[str] = "") -> RenderableType:
    data = prepare_dict(data)
    table = mapping_view_table()
    cols = []
    for key, content in data.items():
        if not content:
            continue

        content = flexitable(content, key)
        if isinstance(content, ConsoleRenderable) and not isinstance(content, Markdown):
            cols.append(border_panel(content, title=flexitable(key)))
            # table.add_row(key, border_panel(content))
        else:
            table.add_row(key, content)

    cols.insert(0, table)
    if header:
        return Columns(cols)

    # cols.sort(key=lambda x: console.measure(x).maximum)
    # return new_table(rows=it.zip_longest(*(iter(cols),) * 1))
    table = new_table()
    row, width = [], 0
    rows: List[Panel] = []
    for rend in cols:
        this_width = console.measure(rend).maximum
        if width + this_width > console.width:
            # lines.append(simple_panel(new_table(rows=[row], padding=(0, 0))))
            rows.append(simple_panel(new_table(rows=[row], padding=(0, 0))))
            row, width = [rend], this_width
        else:
            row.append(rend)
            width += this_width
    rows.append(simple_panel(new_table(rows=[row], padding=(0, 0))))
    table.add_rows([[r] for r in rows])
    return table


list_table = partial(new_table, expand=False, box=box.SIMPLE_HEAD, border_style="cyan")


@flexitable.register
def _(data: List[str], header: str = "") -> RenderableType:
    call = FIELDS_MAP.get(header)
    return call("\n".join(data)) if call else "\n".join(map(format_with_color, data))


@flexitable.register
def _(data: List[int], header: Optional[str] = None) -> RenderableType:
    return border_panel(Columns(str(x) for x in data))


@flexitable.register
def _(data: List[JSONDict], header: Optional[str] = None) -> RenderableType:
    data = [prepare_dict(item) for item in data]
    all_keys = dict.fromkeys(it.chain.from_iterable(tuple(d.keys()) for d in data))
    keys = {k: None for k in all_keys if any((d.get(k) for d in data))}.keys()

    if len(keys) > 17:
        table = list_table(show_header=False)
        for item in data:
            table.add_row(flexitable({k: v for k, v in item.items() if k in keys}))
        return table

    overlap = set(map(type, data[0].values())) & {int, float, str}
    keysstr = " ".join(keys)
    counting = any(x in keysstr for x in ["count", "sum_", "duration"])
    if len(overlap) == 2 and counting:
        return counts_table(data, header=header or "")

    table = list_table(show_header=True)
    for key in keys:
        table.add_column(key)
    for item in data:
        table.add_dict_item(item, transform=flexitable)
    for col, old in ((c, str(c.header)) for c in table.columns if c):
        new = DISPLAY_HEADER.get(old) or old
        col.header = wrap(new, f"{predictably_random_color(new)}")

    return table


@flexitable.register
def _(data: List[Any], header: Optional[str] = None) -> ConsoleRenderable:
    table = list_table(show_header=False)
    for item in data:
        content = flexitable(item)
        if isinstance(content, Iterable) and not isinstance(content, str):
            table.add_row(*content)
        else:
            table.add_row(content)

    return simple_panel(table)



import random
import re
import time
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from difflib import SequenceMatcher
from functools import lru_cache, partial
from itertools import islice
from math import copysign
from os import environ, path
from string import ascii_letters, ascii_lowercase, digits, printable, whitespace
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    SupportsFloat,
    Type,
    Union,
)

from rich import box
from rich.align import Align
from rich.bar import Bar
from rich.console import Console, ConsoleRenderable, RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from rich.theme import Theme
from rich.tree import Tree

JSONDict = Dict[str, Any]
SPLIT_PAT = re.compile(r"[;,] ?")


def wrap(text: str, tag: str) -> str:
    return f"[{tag}]{text}[/]"


def format_new(string: str) -> str:
    return wrap(re.sub(r"(^\s+$)", "[u green]\\1[/]", string), "b green")


def format_old(string: str) -> str:
    return wrap(string, "b s red")


def fmtdiff(change: str, before: str, after: str) -> str:
    if change == "insert":
        return format_new(after)
    elif change == "delete":
        return format_old(before)
    elif change == "replace":
        return format_old(before) + format_new(after)
    else:
        return wrap(before, "dim")


def make_difftext(
    before: str, after: str, junk: str = whitespace + ascii_letters
) -> str:
    before = re.sub(r"\\?\[", r"\\[", before)
    after = re.sub(r"\\?\[", r"\\[", after)

    matcher = SequenceMatcher(lambda x: x in junk, autojunk=False, a=before, b=after)
    diff = ""
    for code, a1, a2, b1, b2 in matcher.get_opcodes():
        diff = diff + (fmtdiff(code, before[a1:a2], after[b1:b2]) or "")
    return diff


def duration2human(duration: SupportsFloat, acc: int = 1) -> str:
    diff = timedelta(seconds=float(duration))
    return ":".join(
        map(
            lambda x: str(x).zfill(2),
            [
                diff.days * 24 + diff.seconds // 3600,
                diff.seconds % 3600 // 60,
                diff.seconds % 60,
            ],
        )
    )


def fmt_time(seconds: int) -> Iterable[str]:
    abs_seconds = abs(seconds)
    return (
        "{:>3}{}".format(int(copysign(num, seconds)), unit)
        for num, unit in (
            (abs_seconds // 86400, "d"),
            (abs_seconds // 3600, "h"),
            (abs_seconds % 3600 // 60, "m"),
            (abs_seconds % 60, "s"),
        )
        if num
    )


def get_theme() -> Optional[Theme]:
    _path = path.join(
        environ.get("XDG_CONFIG_HOME") or path.expanduser("~/.config"),
        "rich",
        "config.ini",
    )
    if path.exists(_path):
        return Theme.read(_path)
    return None


def make_console(**kwargs: Any) -> Console:
    return Console(
        theme=get_theme(),
        force_terminal=True,
        force_interactive=True,
        emoji=True,
        **kwargs,
    )


class NewTable(Table):
    def __init__(self, *args: str, **kwargs: Any) -> None:
        ckwargs = dict(
            overflow=kwargs.pop("overflow", "fold"),
            justify=kwargs.pop("justify", "left"),
            vertical=kwargs.pop("vertical", "middle"),
        )
        super().__init__(**kwargs)
        for idx, arg in enumerate(args):
            self.add_column(arg, **ckwargs)

    def add_rows(self, rows: Iterable[Iterable[RenderableType]]) -> None:
        """Add multiple rows to the table."""
        for row in rows:
            self.add_row(*row)

    @property
    def colnames(self) -> List[str]:
        """Provide a mapping between columns names / ids and columns."""
        return [str(c.header) for c in self.columns]

    @property
    def colmap(self) -> Dict[str, int]:
        """Provide a mapping between columns names / ids and columns."""
        return {str(c.header): c._index for c in self.columns if c.header}

    def add_dict_item(
        self,
        item: JSONDict,
        transform: Callable[[Any, str], Any] = lambda x, y: x,
        **kwargs: Any,
    ) -> None:
        """Take the required columns / keys from the given dictionary item."""
        vals = (transform(item.get(c), c) for c in self.colnames)
        self.add_row(*vals, **kwargs)


def new_table(*headers: str, **kwargs: Any) -> NewTable:
    default = dict(
        border_style="black",
        show_edge=False,
        show_header=False,
        pad_edge=False,
        highlight=True,
        row_styles=["white"],
        expand=False,
        title_justify="left",
    )
    if headers:
        default.update(
            header_style="bold misty_rose1", box=box.SIMPLE_HEAVY, show_header=True
        )
    rows = kwargs.pop("rows", [])
    table = NewTable(*headers, **{**default, **kwargs})
    if rows:
        table.add_rows(rows)
    return table


@lru_cache(None)
def predictably_random_color(string: str) -> str:
    random.seed(string)
    rand = partial(random.randint, 60, 200)
    return "#{:02X}{:02X}{:02X}".format(rand(), rand(), rand())


def format_with_color(name: str) -> str:
    return wrap(name, f"b {predictably_random_color(name)}")


def simple_panel(content: RenderableType, **kwargs: Any) -> Panel:
    default: JSONDict = dict(
        title_align="left", subtitle_align="right", box=box.SIMPLE, expand=False
    )
    if "title" in kwargs:
        kwargs["title"] = wrap(kwargs["title"], "b")
    if kwargs.pop("align", "") == "center":
        content = Align.center(content, vertical="middle")
    return Panel(content, **{**default, **kwargs})


def border_panel(content: RenderableType, **kwargs: Any) -> Panel:
    return simple_panel(
        content, **{**dict(box=box.ROUNDED, border_style="dim"), **kwargs}
    )


def md_panel(content: str, **kwargs: Any) -> Panel:
    return Markdown(
        content,
        justify="left",
        # inline_code_lexer="python",
        # inline_code_theme="paraiso-dark"
        # re.sub(
        #     r"```\n",
        #     "```python\n",
        #     content.replace("suggestion", "python"),
        #     count=1,
        # )
        # content.replace("suggestion", "python")
    )


def new_tree(
    values: Iterable[ConsoleRenderable] = [], title: str = "", **kwargs
) -> Tree:
    color = predictably_random_color(title or str(values))
    default: JSONDict = dict(guide_style=color)
    tree = Tree(wrap(title, "b"), **{**default, **kwargs})

    for val in values:
        tree.add(val)
    return tree


def get_country(code: str) -> str:
    return format_with_color(code)
    # try:
    #     # country = (
    #     #     pycountry.countries.lookup(code)
    #     #     .name.replace("Russian Federation", "Russia")
    #     #     .replace("Czechia", "Czech Republic")
    #     #     .replace("North Macedonia", "Macedonia")
    #     #     .replace("Korea, Republic of", "South Korea")
    #     # )
    #     country = "Russia"
    #     return f":flag_for_{country.lower().replace(' ', '_')}: {country}"
    # except LookupError:
    #     return "Worldwide"


def colored_with_bg(string: str) -> str:
    sep = wrap("a", "#000000 on #000000")
    return (
        sep + wrap(string, f"bold {predictably_random_color(string)} on #000000") + sep
    )


def _colored_split(strings: List[str]) -> str:
    return " ".join(map(format_with_color, strings))


def unsorted_colored_split(string: str) -> str:
    return _colored_split(SPLIT_PAT.split(string))


def colored_split(string: str) -> str:
    return _colored_split(sorted(SPLIT_PAT.split(string)))


def progress_bar(
    count: float, total_max: float, item_max: Optional[float] = None
) -> Bar:
    use_max = total_max
    if item_max is not None:
        use_max = item_max
    ratio = count / use_max if use_max else 0
    random.seed(str(total_max))
    rand = partial(random.randint, 50, 180)

    def norm():
        return round(rand() * ratio)

    color = "#{:0>2X}{:0>2X}{:0>2X}".format(norm(), norm(), norm())
    return Bar(use_max, 0, count, color=color)


def _get_val(value: Any, field: str) -> str:
    return FIELDS_MAP[field](value) if value is not None else ""


def get_val(obj: JSONDict, field: str) -> str:
    return _get_val(obj.get(field), field)


def counts_table(data: List[JSONDict], header: str = "") -> Table:
    keys = dict.fromkeys(data[0])
    count_col_name = "count"
    if count_col_name not in keys:
        first = data[0]
        for key in keys:
            if key != "total" and isinstance(first[key], (int, float)):
                count_col_name = key

    all_counts = {float(i.get(count_col_name, 0)) for i in data}
    num_type: Type = float
    if len({c % 1 for c in all_counts}) == 1:
        num_type = int
    total_max = max(all_counts)

    # ensure count_col is at the end
    headers = [k for k in keys if k not in {count_col_name, "total"}]
    table = new_table(*headers, "", count_col_name, overflow="fold", vertical="middle")
    for item in data:
        item_count = float(item.pop(count_col_name, 0))
        item_max = item.pop("total", None)
        if item_max is not None:
            item_max = float(item_max)
            item_table_val = f"{num_type(item_count)}/{num_type(item_max)}"
        elif "duration" in count_col_name:
            item_table_val = duration2human(item_count, 2)
        else:
            item_table_val = str(num_type(item_count))
        table.add_row(
            *map(lambda x: get_val(item, x), headers),
            item_table_val,
            progress_bar(item_count, total_max, item_max=item_max),
        )
    if count_col_name in {"duration", "total_duration"}:
        table.caption = "Total " + duration2human(float(sum(all_counts)), 2)
        table.caption_justify = "left"
    if header:
        table.title = header
    return table


def timestamp2datetime(timestamp: Union[str, int, float, None]) -> datetime:
    if isinstance(timestamp, str):
        timestamp = re.sub("[.]\d+", "", timestamp.strip("'"))
        formats = [
            "%Y-%m-%dT%H:%M:%SZ",
            "%Y%m%dT%H%M%SZ",
            "%Y-%m-%d %H:%M:%S",
            "%Y-%m-%dT%H:%M:%S%z",
        ]
        for fmt in formats:
            try:
                return datetime.strptime(timestamp, fmt)
            except ValueError:
                pass
    return datetime.fromtimestamp(int(float(timestamp or 0)), tz=timezone.utc)


def timestamp2datetimestr(timestamp: Union[str, int, float, None]) -> str:
    return timestamp2datetime(timestamp).strftime("%F %T")


def timestamp2timestr(timestamp: Union[str, int, float, None]) -> str:
    return timestamp2datetime(timestamp).strftime("%T")


def time2human(timestamp: Union[int, str], acc: int = 1) -> str:
    try:
        datetime = timestamp2datetime(timestamp)
    except ValueError:
        return timestamp
    diff = datetime.timestamp() - time.time()
    fmted = " ".join(islice(fmt_time(int(diff)), acc))

    if abs(diff) > 86000:
        strtime = datetime.strftime("%F")
    else:
        strtime = datetime.strftime("%T")

    return "[b {}]{}[/]".format("red" if diff < 0 else "green", fmted) + " " + strtime


FIELDS_MAP: Dict[str, Callable[[str], RenderableType]] = defaultdict(
    lambda: str,
    albumtypes=lambda x: "; ".join(
        map(
            format_with_color,
            {
                "album; compilation": "comp",
                "dj-mix; broadcast": "dj-mix",
                "broadcast; dj-mix": "dj-mix",
            }
            .get(x, x)
            .split("; "),
        )
    ),
    author=colored_with_bg,
    # participants=lambda x: "  ".join(map(colored_with_bg, x)),
    user=format_with_color,
    bodyHTML=md_panel,
    desc=md_panel,
    label=format_with_color,
    labels=lambda x: wrap(
        "    ".join(wrap(y["name"].upper(), f"#{y['color']}") for y in x), "b i"
    )
    if isinstance(x, list)
    else colored_split(x),
    catalognum=format_with_color,
    last_played=time2human,
    avg_last_played=lambda x: time2human(x, acc=2),
    since=lambda x: x
    if isinstance(x, str)
    else datetime.fromtimestamp(float(x)).strftime("%F %H:%M"),
    mtime=time2human,
    dt=lambda x: time2human(x, 5),
    start=time2human,
    end=time2human,
    added=time2human,
    entry=time2human,
    due=time2human,
    created=time2human,
    createdAt=lambda x: x if "[/]" in x else time2human(x),
    modified=time2human,
    updated=time2human,
    wait_per_play=lambda x: wrap(
        " ".join(islice(fmt_time(int(float(x))), 1)), "b green"
    ),
    updatedAt=time2human,
    committedDate=time2human,
    bpm=lambda x: wrap(
        x,
        "green"
        if int(x or 0) < 135
        else "b red"
        if int(x or 0) > 230
        else "red"
        if int(x or 0) > 165
        else "yellow",
    ) if isinstance(x, int) else x,
    style=format_with_color,
    genre=colored_split,
    length=timestamp2timestr,
    tracktotal=lambda x: (wrap("{}", "b cyan") + "/" + wrap("{}", "b cyan")).format(*x)
    if isinstance(x, Iterable) and not isinstance(x, str)
    else str(x),
    country=get_country,
    data_source=format_with_color,
    helicopta=lambda x: ":fire: " if x and int(x) else "",
    hidden=lambda x: ":shit: " if x and int(x) else None,
    keywords=lambda x: " ".join(map(colored_with_bg, colored_split(x).split("  ")))
    if isinstance(x, str)
    else x,
    ingr=lambda x: simple_panel(colored_split(x)),
    content=lambda x: md_panel(x) if isinstance(x, str) else x,
    notes=md_panel,
    text=md_panel,
    instructions=md_panel,
    comment=md_panel,
    comments=lambda x: md_panel(
        x.replace("\n0", "\n* 0").replace("\n[", "\n* ["), title="comments"
    ),
    tags=colored_split,
    released=lambda x: x.replace("-00", "") if isinstance(x, str) else str(x),
    calendar=format_with_color,
    source=format_with_color,
    category=format_with_color,
    categories=colored_split,
    # price=lambda x: colored_with_bg(str(x)),
    interview=md_panel,
    benefits=md_panel,
    primary=lambda x: colored_split(x) if isinstance(x, str) else str(x),
    **{"from": format_with_color},
    to=format_with_color,
    creditText=md_panel,
    duration=lambda x: duration2human(x, 2) if isinstance(x, (int, float)) else x,
    total_duration=lambda x: duration2human(x, 2),
    brand=format_with_color,
    mastering=format_with_color,
    answer=md_panel,
    plays=lambda x: wrap(x, "b green"),
    skips=lambda x: wrap(x, "b red"),
    # name=lambda x: wrap(x, "b"),
    description=md_panel,
    body=md_panel,
    event=format_with_color,
    kind=colored_split,
    type_name=format_with_color,
    table=format_with_color,
    endpoint=format_with_color,
    issuetype=format_with_color,
    priority=format_with_color,
    status=format_with_color,
    key=format_with_color,
    assignee=format_with_color,
    subtask_priority=format_with_color,
    subtask_status=format_with_color,
    subtask_key=format_with_color,
    subtask_assignee=format_with_color,
    epic_priority=format_with_color,
    epic_status=format_with_color,
    epic_key=format_with_color,
    Category=format_with_color,
    Description=format_with_color,
    symbol=format_with_color,
    module=format_with_color,
    code=format_with_color,
    entity=format_with_color,
    new=lambda x: "[b green]:heavy_check_mark:[/]"
    if x
    else "[b red]:cross_mark_button:[/]",
    message=lambda x: border_panel(
        Syntax(
            x,
            "python",
            theme="paraiso-dark",
            background_color="black",
            word_wrap=True,
            indent_guides=True,
        )
    ),
    link=lambda x: {
        "blocks": lambda y: wrap(f" {y} ", "b black on red"),
        "is blocked by": lambda y: wrap(y, "b red"),
    }.get(x, lambda x: x)(x),
    album=format_with_color,
    context=lambda x: Syntax(
        x, "python", theme="paraiso-dark", background_color="black", word_wrap=True
    ),
    python=lambda x: Syntax(
        x, "python", theme="paraiso-dark", background_color="black", word_wrap=True
    ),
    CreatedBy=lambda x: Syntax(
        x.replace(";", "\n"),
        "sh",
        theme="paraiso-dark",
        background_color="black",
        word_wrap=True,
    ),
    sql=lambda x: Syntax(
        x.replace("'", ""),
        "sql",
        theme="paraiso-dark",
        background_color="black",
        word_wrap=True,
    ),
    file=lambda x: "/".join(map(format_with_color, x.split("/"))),
    field=lambda x: ".".join(map(format_with_color, x.split("."))),
    log=lambda x: border_panel(
        Syntax(
            x,
            "python",
            theme="paraiso-dark",
            background_color="black",
            word_wrap=True,
            indent_guides=True,
        )
    ),
    diff=lambda x: Text.from_markup(x) if "[/]" in x else md_panel(x),
)

DISPLAY_HEADER: Dict[str, str] = {
    "track": "#",
    "bpm": "🚀",
    "last_played": ":timer_clock: ",
    "mtime": "updated",
    "data_source": "source",
    "helicopta": "[dark red]:helicopter:[/]",
    "hidden": ":no_entry:",
    "track_alt": ":cd:",
    "catalognum": ":pen: ",
    "plays": "[green]:play_button:[/]",
    "skips": "[red]:stop_button:[/]",
    "albumtypes": "types",
}



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rich_tables']

package_data = \
{'': ['*']}

install_requires = \
['multimethod', 'rgbxy>=0.5', 'rich>=12.3.0']

entry_points = \
{'console_scripts': ['table = rich_tables.table:main']}

setup_kwargs = {
    'name': 'rich-tables',
    'version': '0.1.2',
    'description': 'Ready-made rich tables for various purposes',
    'long_description': '# Rich tables\n\n* [Album](#album)\n* [Calendar](#calendar)\n* [Emails](#emails)\n* [Hue](#hue)\n* [Nested json](#nested_json)\n* [Pr](#pr)\n* [Simple json](#simple_json)\n* [Tasks](#tasks)\n* [Timed](#timed)\n\n\n## Album\n![image](svgs/album.svg)\n## Calendar\n![image](svgs/calendar.svg)\n## Emails\n![image](svgs/emails.svg)\n## Hue\n![image](svgs/hue.svg)\n## Nested json\n![image](svgs/nested_json.svg)\n## Pr\n![image](svgs/pr.svg)\n## Simple json\n![image](svgs/simple_json.svg)\n## Tasks\n![image](svgs/tasks.svg)\n## Timed\n![image](svgs/timed.svg)\n',
    'author': 'Šarūnas Nejus',
    'author_email': 'snejus@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snejus/rich-tables',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.3,<4',
}


setup(**setup_kwargs)
